/*
 * tel.c -- telnet protocol communication module for pow-wow
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netdb.h>
#include <netinet/in.h>
#include <arpa/telnet.h>
#ifndef TELOPT_NAWS
# define TELOPT_NAWS 31
#endif
#include <arpa/inet.h>
#ifndef NEXT
# include <unistd.h>
#endif

#ifdef TERM
# include "client.h"
#endif /* TERM */

#include "powwow.h"
#include "term.h"

int socket(), connect();
void bcopy(), bzero();

#ifndef OSF1
unsigned long inet_addr();
#endif

#ifdef NOBCOPY
# define bcopy(a,b,c) memcpy ((b), (a), (c))
#endif
#ifdef NOBZERO
# define bzero(a,b) memset ((a), 0, (b))
#endif

#ifdef TELOPTS
# define TELOPTSTR(n) ((n) > NTELOPTS ? "unknown" : telopts[n])
#endif

static void dosubopt();
static void sendopt();

static int sending_size = 0;	/* true if we're communicating window size */

/*
 * connect to remote host
 * Warning: some voodoo code here
 */
int doconnect(addr, port)
char *addr; int port;
{
  struct sockaddr_in address;
  struct hostent *host_info;
  int err;
  
  bzero((char *)&address, sizeof(address));
  /*
   * inet_addr has a strange design: It is documented to return -1 for
   * malformed requests, but it is declared to return unsigned long!
   * Anyway, this works.
   */    

#ifndef TERM  
  address.sin_addr.s_addr = inet_addr(addr);
  if(address.sin_addr.s_addr != (unsigned int)-1)
     address.sin_family = AF_INET;
  else
   {
     host_info = gethostbyname(addr);
     if(host_info == 0)
     	syserr("gethostbyname:unknown host");
     bcopy(host_info->h_addr, (char*)&address.sin_addr,
	   host_info->h_length);
     address.sin_family = host_info->h_addrtype;
   }
  address.sin_port = htons(port);
  
  sockfd = socket(address.sin_family, SOCK_STREAM, 0);
  if(sockfd < 0)
     syserr("creating socket");
  
  printf("#Trying %s... ", addr); fflush(stdout);
  while((err = connect(sockfd, (struct sockaddr *)&address,
		       sizeof(address))) < 0 && errno == EINTR);
#else /* term */
  
  if ((sockfd = connect_server(0)) < 0)
   {
     syserr("unable to connect to term server");
   }
  else
   {
     send_command(sockfd, C_PORT, 0, "%s:%d", addr, port);
     printf("Connected to term server...\n");
#ifdef TERM_COMPRESS
     send_command(sockfd, C_COMPRESS, 1, "y");
#endif     
     send_command(sockfd, C_DUMB, 1, 0);
   }
#endif /* term */
  
  if ((err < 0) || sockfd >= MAXFDSCAN) 
   {
     close(sockfd);
     sockfd = -1;
   }
  else
     fputs("connected!\n", stdout);
  return sockfd;
}

/*
 * read a maximum of size bytes from remote host
 * using the telnet protocol. return bytes read.
 */
int read_socket(buffer, maxsize)
char *buffer; int maxsize;
{
    typedef enum {
	NORMAL = 0, GOTIAC, GOTWILL, GOTWONT, GOTDO, GOTDONT, GOTSB, GOTSBIAC
    } telnetstate;
    static telnetstate sock_state[MAXFDSCAN];
    telnetstate state = sock_state[sockfd];
    static unsigned char subopt[MAXSUBOPT];
    static int subchars;
    int i;
    unsigned char *p, *s;
    
    while((i = read(sockfd, buffer, maxsize)) < 0 && errno == EINTR);
    if(i <= 0) {
	if(i == 0) {
	    sock_state[sockfd] = NORMAL;
	    disconnect_sess(NULL);
	    return 0;
	}
	syserr("read from socket");
    }
    /* scan through the buffer and interpret telnet protocol escapes */

    for(s = p = (unsigned char*)buffer; i--; s++) {
	switch(state) {
	case NORMAL:
	default:
	    if(*s != IAC)
		*p++ = *s;
	    else
		state = GOTIAC;
	    break;
	case GOTIAC:
	    switch(*s) {
	    case WILL:
		state = GOTWILL; break;
	    case WONT:
		state = GOTWONT; break;
	    case DO:
		state = GOTDO; break;
	    case DONT:
		state = GOTDONT; break;
	    case SB:		/* BUG (multiple sessions):	*/
		state = GOTSB;	/* there is only one subopt buffer */
		subchars = 0;
		break;
	    case IAC:
		*p++ = IAC;
		break;
	    case GA:
		/* I should handle GA as end-of-prompt marker one day */
	    default:
		/* ignore the rest of the telnet commands */
#ifdef TELOPTS
		printf("[skipped IAC <%d>]\n", *s);
#endif
		state = NORMAL; break;
	    }
	    break;
	
	case GOTWILL:
#ifdef TELOPTS
	    printf("[got WILL %s]\n", TELOPTSTR(*s));
#endif
	    switch(*s) {
	    case TELOPT_ECHO:
		/* host echoes, turn off echo here
		 * but only for main session, since we do not want
		 * subsidiary session password entries to block anything
		 * in the main session
		 */
		if (is_main_sess())
		    linemode |= LM_NOECHO;
		sendopt(DO, *s);
		break;
	    case TELOPT_SGA:
		/* this can't hurt */
		linemode |= LM_CHAR;
		set_spec_keys();
		sendopt(DO, *s);
		break;
	    default:
		/* don't accept other options */
		sendopt(DONT, *s);
		break;
	    }
	    state = NORMAL;
	    break;
	
	case GOTWONT:
#ifdef TELOPTS
	    printf("[got WONT %s]\n", TELOPTSTR(*s));
#endif
	    if(*s == TELOPT_ECHO) {
		/* host no longer echoes, we do it instead */
		linemode &= ~LM_NOECHO;
	    }
	    /* accept any WONT */
	    sendopt(DONT, *s);
	    state = NORMAL;
	    break;
	
	case GOTDO:
#ifdef TELOPTS
	    printf("[got DO %s]\n", TELOPTSTR(*s));
#endif
	    switch(*s) {
	    case TELOPT_SGA:
		linemode |= LM_CHAR;
		set_spec_keys();
		/* FALLTHROUGH */
	    case TELOPT_TTYPE:
		sendopt(WILL, *s);
		break;
	    case TELOPT_NAWS:
		sendopt(WILL, *s);
		sending_size = 1;
		send_term_size();
		break;
	    default:
		/* accept nothing else */
		sendopt(WONT, *s);
		break;
	    }
	    state = NORMAL;
	    break;
	
	case GOTDONT:
#ifdef TELOPTS
	    printf("[got DONT %s]\n", TELOPTSTR(*s));
#endif
	    if(*s == TELOPT_SGA) {
		linemode &= ~LM_CHAR;
		set_spec_keys();
	    }
	    sendopt(WONT, *s);
	    state = NORMAL;
	    break;
	
	case GOTSB:
	    if(*s == IAC) {
		state = GOTSBIAC;
	    } else {
		if(subchars < MAXSUBOPT)
		    subopt[subchars++] = *s;
	    }
	    break;
	
	case GOTSBIAC:
	    if(*s == IAC) {
		if(subchars < MAXSUBOPT)
		    subopt[subchars++] = IAC;
		state = GOTSB;
	    } else if(*s == SE) {
		subopt[subchars] = '\0';
		dosubopt((char*)subopt);
		state = NORMAL;
	    } else {
		/* problem! I haven't the foggiest idea of what to do here.
		 * I'll just ignore it and hope it goes away. */
		printf("#telnet: got IAC <%d> instead of IAC SE!\n", (int)*s);
		state = NORMAL;
	    }
	    break;
	}
    }
    sock_state[sockfd] = state;
    return (char*)p - buffer;
}

/*
 * send an option negotiation
 * 'what' is one of WILL, WONT, DO, DONT
 */
static void sendopt(what, opt)
int what; unsigned opt;
{
   static unsigned char buf[3] = { IAC, 0, 0 };
   buf[1] = what; buf[2] = opt;
   if(write(sockfd, buf, 3) < 3)
     syserr("write option");
#ifdef TELOPTS
   printf("[sent %s %s]\n", (what == WILL) ? "WILL" :
	  (what == WONT) ? "WONT" :
	  (what == DO) ? "DO" : (what == DONT) ? "DONT" : "error",
	  TELOPTSTR(opt));
#endif
}

/*
 * Send current terminal size (RFC 1073)
 */
void send_term_size()
{
    if(sending_size) {
	static unsigned char buf[9] =
	    { IAC, SB, TELOPT_NAWS, 0, 0, 0, 0, IAC, SE };

	buf[3] = cols >> 8;
	buf[4] = cols & 0xff;
	buf[5] = lines >> 8;
	buf[6] = lines & 0xff;
	if(write(main_sess(), buf, 9) < 9)
	    syserr("write term size");
#ifdef TELTOPS
	printf("[term size %d %d]\n", cols, lines);
#endif
    }
}

/*
 * process suboptions.
 * so far, only terminal type is processed but future extensions are
 * window size, X display location, etc.
 */
static void dosubopt(str)
char *str;
{
    char buf[256], *term;
  
    if(str[0] == TELOPT_TTYPE) {
	if(str[1] == 1) {
	    /* 1 == SEND */
#ifdef TELOPTS
	    printf("[got SB TERMINAL TYPE SEND]\n");
#endif
	    if(!(term = getenv("TERM"))) term = "unknown";
	    sprintf(buf, "%c%c%c%c%s%c%c", IAC, SB, TELOPT_TTYPE, 0,
		    term, IAC, SE);	/* 0 == IS */
	    if(write(sockfd, buf, strlen(term) + 6) < strlen(term) + 6)
		syserr("write subopt to socket");
#ifdef TELOPTS
	    printf("[sent SB TERMINAL TYPE IS %s]\n", term);
#endif
	}
    }
}


static char output_buffer[OUTBUFSIZE];
static int output_len = 0;	/* number of characters in output_buffer */
static int output_socket = -1;	/* to which socket buffer should be sent*/

/*
 * put data in the output buffer for transmission to the remote host
 */
void send_to_host(data)
char *data;
{
    int len, space;
    sent += len = strlen(data);
    internal = -1;

    if (!(linemode & LM_NOECHO) && capfile)
        fprintf(capfile, "%s\n", data);

    if(sockfd != output_socket) { /* is there data to another socket? */
	if(output_socket >= 0)
	    flush_output_buffer(); /* then flush it */
	output_socket = sockfd;
    }
    space = OUTBUFSIZE - output_len;
    if(space <= len) {  		/* not enough room? */
	bcopy(data, output_buffer + output_len, space);
	output_len = OUTBUFSIZE;
	flush_output_buffer();
	send_to_host(data + space);
    } else {
        data[len] = '\n';
	bcopy(data, output_buffer + output_len, len + 1);
	output_len += len + 1;
        data[len] = '\0';
    }

}

/*
 * send all buffered data to the remote host
 */
void flush_output_buffer()
{
    int n;
    char *p = output_buffer;

    if (output_len && output_socket == -1) {
      clearinpline(1);
      PRINT_INT ("%s#No open sessions. Use '#connect main <address> <port>' to start a session.\n%s", edattrend, edattrbeg);
      output_len = 0;
      showprompt();
      redraw_input();
      return;
    }
      
    while(output_len) {
	if((n = write(output_socket, p, output_len)) < 0)
	    syserr("write to socket");
	p += n;
	output_len -= n;
    }
}
